import subprocess
import sys
import config
import optparse
import notify
import optparse
import re
import syslog
import logging

# Set python cache for rrdtool
import os
os.environ['PYTHON_EGG_CACHE'] = '/usr/local/nagiosna/tmp'

import rrdtool

logger = logging.getLogger('backend')

class UnknownIndexException(Exception):
    pass

def parse_args():
    
    usage = "usage: %prog -f <filename> -s <sourceid>"
    parser = optparse.OptionParser(usage=usage)
    
    required = optparse.OptionGroup(parser, "Required Options")
    required.add_option('-f', '--filename', help='Specify nfcapd file to run check on', default=None)
    required.add_option('-s', '--sourceid', help='Source ID in the MySQL database.', default=None)
    parser.add_option_group(required)
    
    options, _ = parser.parse_args()
    
    if not options.filename:
        print('Filename is required, cannot continue')
        sys.exit(3)
    
    if not options.sourceid:
        print('Source ID is required, cannot continue.')
        sys.exit(3)
    
    return options

def is_within_range(nagstring, value):
    if not nagstring:
        return False
    import re
    
    value = float(value) 
    first_float = r'(?P<first>(-?[0-9]+(\.[0-9]+)?))'
    second_float= r'(?P<second>(-?[0-9]+(\.[0-9]+)?))'
    actions = [ (r'^%s$' % first_float,lambda y: (value > float(y.group('first'))) or (value < 0)),
                (r'^%s:$' % first_float,lambda y: value < float(y.group('first'))),
                (r'^~:%s$' % first_float,lambda y: value > float(y.group('first'))),
                (r'^%s:%s$' % (first_float,second_float), lambda y: (value < float(y.group('first'))) or (value > float(y.group('second')))),
                (r'^@%s:%s$' % (first_float,second_float), lambda y: not((value < float(y.group('first'))) or (value > float(y.group('second')))))]
    for regstr,func in actions:
        res = re.match(regstr,nagstring)
        if res:
            return func(res)
    raise Exception('Improper warning/critical format.')

def get_relevant_checks(sourceid):
    global logger
    GROUPS = '''
SELECT
    `gid`
FROM
    nagiosna_SGLinker
WHERE
    `sid` = '%s'
''' % sourceid
    group_list = [[-1]] + list(config.query_database(GROUPS))
    group_str = ','.join([str(x[0]) for x in group_list])
    
    QUERY = '''
SELECT 
    nagiosna_Checks.name, `cid`, `warning`, `critical`, `metric`, `rawquery`, nagiosna_Sources.name, nagiosna_Sources.sid, `vid`, nagiosna_Checks.gid, nagiosna_SourceGroups.name
FROM 
    nagiosna_Checks
LEFT JOIN
    nagiosna_Sources
ON
    nagiosna_Sources.sid = nagiosna_Checks.sid
LEFT JOIN
    nagiosna_SourceGroups
ON
    nagiosna_SourceGroups.gid = nagiosna_Checks.gid
WHERE
    (nagiosna_Checks.sid = '%s' OR nagiosna_Checks.gid IN (%s))
    AND 
    `active` = '1'
''' % (sourceid, group_str)
    return config.query_database(QUERY)

def get_aggregate(rawquery, capdfile, metric):
    global logger
    nfdump = [  '/usr/local/bin/nfdump',
                '-r', capdfile,
                '-o', 'csv',
                '-s', 'srcip',
                rawquery ]
    logger.debug('Running nfdump - ' + ' '.join(nfdump))
    proc = subprocess.Popen(nfdump, stdout=subprocess.PIPE)
    proc.wait()

    stdout = proc.stdout.readlines()
    summary = stdout[-1].decode('utf-8').rstrip().split(',')

    if metric == 'flows':
        index = 0
    elif metric == 'bytes':
        index = 1
    elif metric == 'packets':
        index = 2
    elif metric == 'bps':
        index = 3
    elif metric == 'pps':
        index = 4
    elif metric == 'bpp':
        index = 5
    else:
        index = 0
    
    try:
        retval = summary[index]
    except IndexError:
        raise UnknownIndexException
    return retval

def make_output(source, rawquery, metric, result, warning, critical):
    if rawquery:
        stdout = "%s on %s with filter [%s] is %s" % (metric, source, rawquery, str(result))
    else:
        stdout = "%s on %s is %s" % (metric, source, str(result))
    perfdata = "%s=%s;%s;%s;0" % (metric, result, warning, critical)
    return "%s | %s" % (stdout, perfdata)

def update_check_db(checkid, stdout, rawval, returncode):
    QUERY = '''
UPDATE
    `nagiosna_Checks`
SET
    `lastcode` = %s,
    `laststdout` = '%s',
    `lastrun` = NOW(),
    `lastval` = %s
WHERE
    `cid` = %s;''' % (str(returncode), str(stdout), str(rawval), str(checkid))
    syslog.syslog(QUERY)
    config.query_database(QUERY, commit=True)

def get_group_capd_files(gid, filename):
    filepath = '%%s/flows/%s' % os.path.basename(filename)
    QUERY = '''
SELECT
    nagiosna_Sources.directory
FROM
    nagiosna_SGLinker
JOIN
    nagiosna_Sources
ON
    nagiosna_SGLinker.sid = nagiosna_Sources.sid
WHERE
    nagiosna_SGLinker.gid = '%s'
''' % gid
    
    return [filepath % x[0] for x in config.query_database(QUERY)]

def get_view_capd_file(sid, vid, filename):
    filepath = '%%s/views/%%s/%s' % os.path.basename(filename)
    QUERY = '''
SELECT
    nagiosna_Sources.directory, nagiosna_Views.name
FROM
    nagiosna_Sources
JOIN
    nagiosna_SourcesViewsLinker
ON
    nagiosna_SourcesViewsLinker.sid = nagiosna_Sources.sid
JOIN
    nagiosna_Views
ON
    nagiosna_SourcesViewsLinker.vid = nagiosna_Views.vid
WHERE
    nagiosna_Views.vid = '%s' AND nagiosna_Sources.sid = '%s'
''' % (vid, sid)
    sourcepath, viewname = config.query_database(QUERY)[0]
    safe_viewname = re.sub('[^a-zA-Z0-9-_\.]', '', viewname)
    return filepath % (sourcepath, safe_viewname), viewname

def get_rrd_paths(sid):

    base_bandwidth = 'bandwidth.rrd'

    (directory, name) = get_source_directory(sid)

    new_path = os.path.join(directory, base_bandwidth)

    return new_path

def get_source_directory(source_id):

    query = '''
SELECT 
    nagiosna_Sources.directory,
    nagiosna_Sources.name
FROM 
    nagiosna_Sources
WHERE
    nagiosna_Sources.sid = %s
    ''' % (source_id)

    value = config.query_database(query)

    # In testing this was returning a 1-tuple that contained a 2-tuple with our values
    # So we'll get our desired 2-tuple like this.
    if len(value) == 1:
        value = value[0]

    return value

def get_failures(sid):

    bandwidth_path = get_rrd_paths(sid)

    columns = ['bytes', 'flows', 'packets']
    rrd_args = ['--json', '--start', '-600', '--end', '-300']


    for index, col in enumerate(columns):
        rrd_args.append(make_defs(col, bandwidth_path, '', 'FAILURES'))
        rrd_args.append(make_xport(col, index))

    result = rrdtool.xport(*rrd_args)

    # If we don't get data back there was something wrong with our query
    if 'data' not in result:
        return None

    # Set result to a list with one entry
    result = result['data']
    # Get the only entry (a 3-tuple)
    result = result[0]
    # Reduce this to just 1 float
    result = float(sum(result))
    return result

def make_defs(metric, rrd, metric_prefix='', round_robin_archive='AVERAGE'):
    ret = [
        'DEF',
        '%s%s=%s' % (metric_prefix, metric, rrd),
        metric,
        round_robin_archive
    ]
    return ':'.join(ret)

def make_xport(key, label):
    ret = [
        'XPORT',
        key,
        "'%s'" % (label)
    ]
    return ':'.join(ret)

def main(options=None):
    
    global logger
    
    if options is None:
        options = parse_args()
    
    checks = get_relevant_checks(options.sourceid)
    
    for check in checks:
        name, checkid, warning, critical, metric, rawquery, sourcename, sid, vid, gid, groupname = check
        
        try:

            if metric == 'abnormal_behavior':

                metric += '_events'
                if gid or vid:
                    stdout = "Abnormal behavior checks can only be run on sources, NOT on views or sourcegroups"
                    rawval = 0
                    returncode = 3
                    update_check_db(checkid, stdout, rawval, returncode)
                    return
                # lets be extra safe
                elif sid:
                    title = sourcename

                warning = '0'
                critical = '0'

                rawval = get_failures(sid)

            elif gid:
                files = get_group_capd_files(gid, options.filename)
                rawval = 0
                for f in files:
                    rawval += float(get_aggregate(rawquery, f, metric))
                title = groupname
            elif vid != 0 and vid != None:
                filename, viewname = get_view_capd_file(sid, vid, options.filename)
                rawval = get_aggregate(rawquery, filename, metric)
                title = "%s view %s" % (sourcename, viewname)
            elif sid:
                rawval = get_aggregate(rawquery, options.filename, metric)
                title = sourcename
        except UnknownIndexException:
                stdout = 'Unable to query data from source. No flow data found.'
                update_check_db(checkid, stdout, 0, 3)
                try:
                    notify.send_notifications(sourcename, checkid, metric, rawquery, 3, stdout)
                except Exception as e:
                    logger.exception(e)
                continue
        
        
        stdout = make_output(title, rawquery, metric, rawval, warning, critical)
        
        try:
            if is_within_range(critical, rawval):
                returncode = 2
            elif is_within_range(warning, rawval):
                returncode = 1
            else:
                returncode = 0
        except Exception as e:
            returncode = 3
        
        
        update_check_db(checkid, stdout, rawval, returncode)

        # Continue running checks if error in a check above
        try:
            notify.send_notifications(sourcename, checkid, metric, rawquery, returncode, stdout)
        except Exception as e:
            logger.exception(e)
            continue

if __name__ == '__main__':
    try:
        main()
        logger.info('Successfully ran check...')
    except Exception as e:
        logger.exception(e)
        sys.exit(3)
        
