import smtplib
from email.mime.text import MIMEText
import config
import xml.etree.ElementTree as ET
import logging
import os
import subprocess

# Make this work for both python2/3

try:
    import urllib.request
    import urllib.parse
except ImportError:
    import urllib2
    import urllib

try:
    urlencode = urllib.parse.urlencode
except AttributeError:
    urlencode = urllib.urlencode

try:
    urlrequest = urllib.request.Request
except AttributeError:
    urlrequest = urllib2.Request

try:
    urlopen = urllib.request.urlopen
except AttributeError:
    urlopen = urllib2.urlopen


logger = logging.getLogger('backend')

def send_email(recipient, subject, body):
    FROM = 'nagiosna-servbot@localhost'
    msg = MIMEText(body)
    msg['Subject'] = subject
    msg['From'] = FROM
    msg['To'] = recipient
    
    s = smtplib.SMTP('localhost')
    s.sendmail(FROM, [recipient], msg.as_string())
    s.quit()

def send_traps(sourcename, checkid, returncode, stdout):
    global logger
    QUERY = '''
SELECT 
    `tid`
FROM
    nagiosna_ChecksTrapsLinker
WHERE
    `cid` = '%s';
''' % checkid
    devices = config.query_database(QUERY)
    
    SNMP_TMPL = '''
SELECT
    *
FROM 
    nagiosna_TrapReceivers
WHERE
    `tid` = '%s';'''
    
    for device in devices:
        info = config.query_database(SNMP_TMPL % device)
        
        try:
            d = info[0]
        except IndexError:
            logger.error('No such device: tid = %s' % device)
            continue
        
        try:
            snmp = build_snmp_trap_cmd(d, sourcename, returncode, stdout)
            subprocess.Popen(snmp)
        except Exception:
            logger.exception('Could not send trap')
            continue

def build_snmp_trap_cmd(snmp_info, sourcename, returncode, stdout):
    
    cmd = ['/usr/bin/snmptrap']
    
    (   tid, 
        name, 
        version, 
        community, 
        authlevel, 
        authprotocol, 
        authpassword, 
        privprotocol, 
        privpassword, 
        username, 
        port, 
        ip              ) = snmp_info
    
    if version == 2:
        cmd += ['-v2c']
        cmd += ['-c', community]
        
    else:
        cmd += ['-v3']
        cmd += ['-u', username]
        if authlevel == 'authPriv':
            cmd += ['-a', authprotocol]
            cmd += ['-A', authpassword]
            cmd += ['-x', privprotocol]
            cmd += ['-X', privpassword]
        if authlevel == 'authNoPriv':
            cmd += ['-a', authprotocol]
            cmd += ['-A', authpassword]
    
    cmd += ['%s:%s' % (ip, port)]
    
    cmd += ['']
    cmd += ['NAGIOS-NOTIFY-MIB::nSvcEvent', 'nSvcHostname', 's', 'NagiosNA']
    cmd += ['nSvcDesc', 's', '%s' % str(sourcename)]
    cmd += ['nSvcStateID', 'i', str(returncode)]
    cmd += ['nSvcOutput', 's', '%s' % str(stdout)]
    
    return cmd

def get_user_emails(checkid):
    QUERY = '''
SELECT 
    `email`, `first_name`, `last_name`
FROM 
    nagiosna_users 
JOIN 
    nagiosna_ChecksUsersLinker 
ON 
    nagiosna_ChecksUsersLinker.uid = id 
WHERE 
    nagiosna_ChecksUsersLinker.cid = '%s';''' % str(checkid)
    results = config.query_database(QUERY)
    return results

def get_nrdp_info(checkid):
    QUERY = '''
SELECT 
    `hostname`, `servicename`, `nrdp`, `token` 
FROM 
    nagiosna_ChecksHSALinker hsal 
JOIN 
    nagiosna_HostServiceAssociations hsa 
ON 
    hsal.aid = hsa.id 
JOIN 
    nagiosna_NagiosServers ns 
ON 
    hsa.serverid = ns.id
WHERE
    hsal.cid = '%s';''' % str(checkid)
    results = config.query_database(QUERY)
    return results
    
def make_nrdp_xml(hostname, servicename, returncode, stdout):
    checkresults = ET.Element('checkresults')
    checkresult = ET.SubElement(checkresults, 'checkresult', type='service')
    xhostname = ET.SubElement(checkresult, 'hostname')
    xhostname.text = hostname
    xservicename = ET.SubElement(checkresult, 'servicename')
    xservicename.text = servicename
    state = ET.SubElement(checkresult, 'state')
    state.text = str(returncode)
    output = ET.SubElement(checkresult, 'output')
    output.text = stdout
    return ET.tostring(checkresults)

def post_xml_to_nrdp(url, xml, token):
    global logger
    values = {  'XMLDATA': xml,
                'token': token,
                'cmd': 'submitcheck' }

    data = urlencode(values)
    req = urlrequest(url, data)

    # Log the XML data in debug log
    logger.debug('URL: %s' % url)
    logger.debug('Data: %s' % data)

    response = urlopen(req)
    logger.debug(response.read())

def send_nrdp_notifications(checkid, stdout, returncode):
    global logger
    nrdp_info = get_nrdp_info(checkid)
    for hostname, servicename, nrdp, token in nrdp_info:
        xml = make_nrdp_xml(hostname, servicename, returncode, stdout)
        try:
            post_xml_to_nrdp(nrdp, xml, token)
        except Exception:
            logger.exception('Could not send to NRDP server: %s' % hostname)
            continue

def send_emails(sourcename, checkid, metric, rawquery, returncode, stdout):
    global logger
    contact_info = get_user_emails(checkid)
    body = '''
%%s,

The source %s has been evaluated for flows given the following query:

%s

Looking at the number results (%s) the check has returned as %s

Check output:

%s

- Nagios Network Analyzer''' % (sourcename, rawquery, metric, get_state(returncode), stdout)
    subject = '%s: %s on %s' % (get_state(returncode), metric, sourcename)
    for email, firstname, lastname in contact_info:
        name = "%s %s" % (firstname, lastname)
        specific = body % name

        # Try to send email
        try:
            send_email(email, subject, specific)
        except Exception:
            logger.exception('Could not send email to %s' % email)
            continue

def run_local_cmds(sourcename, checkid, returncode, stdout):
    global logger
    # Get the commands for the current check...
    QUERY = '''SELECT *
FROM nagiosna_ChecksCmdLinker cl 
JOIN nagiosna_Commands c ON cl.cmdid = c.id
WHERE cl.cid = '%s';''' % str(checkid)
    commands = config.query_database(QUERY)

    # If there are commands, then run them
    for command in commands:
        location = command[4]
        script = command[5]
        arguments = command[6]

        d = { '%sourcename%': sourcename,
              '%state%': get_state(returncode),
              '%returncode%': str(returncode),
              '%output%': stdout }

        for k, v in d.iteritems():
            arguments = arguments.replace(k, v)

        full_path = "%s/%s" % (location, script)
        full_path = full_path.replace('//', '/');
        cmd = "%s %s" % (full_path, arguments)

        try:
            os.system(cmd)
        except Exception:
            logger.exception('Could not run command: %s' % cmd)
            continue

def send_notifications(sourcename, checkid, metric, rawquery, returncode, stdout):
    send_emails(sourcename, checkid, metric, rawquery, returncode, stdout)
    send_nrdp_notifications(checkid, stdout, returncode)
    send_traps(sourcename, checkid, returncode, stdout)
    run_local_cmds(sourcename, checkid, returncode, stdout)

def get_state(returncode):
    if returncode == 0:
        return 'OK'
    elif returncode == 1:
        return 'WARNING'
    elif returncode == 2:
        return 'CRITICAL'
    else:
        return 'UNKNOWN'